# Copyright 2002-2003 Andrew Loewenstern, All Rights Reserved
# See khashmir.py for license information
# Updated and modified for ABC_OKC : Old King Cole

import sys

from BitTornado.bencode import bencode, bdecode
from BitTornado.CurrentRateMeasure import Measure
from defer import Deferred
from traceback import format_exception
from expiringdict import ExpiringDict
from actions import Expirer
from const import KRPC_TIMEOUT, KRPC_AGE, KRPC_DELAY

# commands
TID = 't'
REQ = 'q'
RSP = 'r'
TYP = 'y'
ARG = 'a'
ERR = 'e'

# errors
KRPC_ERROR = 1
KRPC_ERROR_METHOD_UNKNOWN = 2
KRPC_ERROR_RECEIVED_UNKNOWN = 3
KRPC_ERROR_TIMEOUT = 4
KRPC_SOCKET_ERROR = 5


class Hostbroker:       
    def __init__(self, server, addr, sock, sched):
        self.server = server
        self.addr = addr
        self.sock = sock
        self.sched = sched
        self.connections = ExpiringDict()
        Expirer(self.connections, self.sched, KRPC_AGE, KRPC_DELAY, KRPC_DELAY)
        self.uprate = Measure(10)

    def connectionForAddr(self, addr):
        conn = self.connections.get(addr)
        if conn:
            self.connections.touch(addr)
        else:
            conn = KRPC(addr, self.server, self.sock, self.sched, self)
            self.connections[addr] = conn
        return conn

    def data_came_in(self, addr, datagram):
        c = self.connectionForAddr(addr)
        c.datagramReceived(datagram)

    def connection_lost(self, socket):
        pass


# Connection
class KRPC:
    __slots__ = ('noisy', 'addr', 'server', 'sock', 'sched', 'udp', 'tids', 'mtid')
    noisy = 0
    def __init__(self, addr, server, sock, sched, udp):
        self.addr = addr
        self.server = server
        self.sock = sock
        self.sched = sched
        self.udp = udp
        self.tids = {}
        self.mtid = 0

    def datagramReceived(self, str):
        # bdecode
        try:
            msg = bdecode(str)
        except Exception, e:
            if self.noisy:
                print "Response decode error : " + `e`
        else:
            #print 'RECEIVED MESSAGE FROM', self.addr, ':', msg
            # look at msg type
            if msg[TYP] == REQ:
                # if request
                # tell server to handle
                f = getattr(self.server, "krpc_" + msg[REQ], None)
                msg[ARG]['_krpc_sender'] = self.addr
                if f and callable(f):
                    try:
                        ret = apply(f, (), msg[ARG])
                    except Exception, e:
                        # error
                        resp = {TID: msg[TID], TYP: ERR, ERR: `format_exception(type(e), e, sys.exc_info()[2])`}
                    else:
                        # response
                        resp = {TID: msg[TID], TYP: RSP, RSP: ret}
                    # send
                else:
                    # unknown method
                    if self.noisy:
                        print "Don't know about method : %s" % msg[REQ]
                    resp = {TID: msg[TID], TYP: ERR, ERR: KRPC_ERROR_METHOD_UNKNOWN}
                #print 'RESPONDING TO', self.addr, ':', resp
                encodedresp = bencode(resp)
                try:
                    self.sock.sendto(encodedresp, 0, self.addr)
                except:
                    #print ">>>>>> KRPC_SOCKET_ERROR"
                    pass
                else:
                    self.udp.uprate.update_rate(len(encodedresp))
            elif msg[TYP] == RSP:
                # if response
                if self.tids.has_key(msg[TID]):
                    df = self.tids[msg[TID]]
                    # callback
                    df.callback({'rsp': msg[RSP], '_krpc_sender': self.addr})
                    del self.tids[msg[TID]]
                #else:
                    # no tid, this transaction timed out already...
                    #print 'Timeout : ' + `msg[RSP]['id']`
            elif msg[TYP] == ERR:
                # if error
                if self.tids.has_key(msg[TID]):
                    df = self.tids[msg[TID]]
                    # 	callback
                    df.errback(msg[ERR])
                    del self.tids[msg[TID]]
            else:
                # unknown message type
                #print "Unknown message type : " + `msg`
                if self.tids.has_key(msg[TID]):
                    df = self.tids[msg[TID]]
                    # callback
                    df.errback(KRPC_ERROR_RECEIVED_UNKNOWN)
                    del self.tids[msg[TID]]

    def timeOut(self, id):
        if self.tids.has_key(id):
            df = self.tids[id]
            #print ">>>>>> KRPC_ERROR_TIMEOUT"
            df.errback(KRPC_ERROR_TIMEOUT)
            del self.tids[id]

    def sendRequest(self, method, args):
        # make message
        # send it
        msg = {TID: chr(self.mtid), TYP: REQ, REQ: method, ARG: args}
        self.mtid = (self.mtid + 1) % 256
        d = Deferred()
        self.tids[msg[TID]] = d
        self.sched(self.send, 0, [msg])
        return d
 
    def send(self, msg):
        mtid = msg[TID]
        self.sched(self.timeOut, KRPC_TIMEOUT, [mtid])
        #print 'SENDING REQUEST TO', self.addr, ':', msg
        encodedmsg = bencode(msg)
        try:
            self.sock.sendto(encodedmsg, 0, self.addr)
        except:
            df = self.tids[mtid]
            #print ">>>>>> KRPC_SOCKET_ERROR"
            df.errback(KRPC_SOCKET_ERROR)
            del self.tids[mtid]
        else:
            self.udp.uprate.update_rate(len(encodedmsg))
